import Foundation
import SwiftUI

final class BluetoothClient: ObservableObject {
    
    private let updateInterval = 3.0

    private let minimumOffsetToUpdateDeviceClock = 5.0
    
    private let connection = DeviceManager()

    private let storage: TemperatureStorage
    
    init(storage: TemperatureStorage, deviceInfo: DeviceInfo? = nil) {
        self.storage = storage
        self.deviceInfo = deviceInfo
        connection.delegate = self
    }
    
    func connect() -> Bool {
        connection.connect()
    }
    
    @Published
    private(set) var deviceState: DeviceState = .disconnected {
        didSet {
            print("State: \(deviceState.text)")
            if case .configured = deviceState {
                startRegularUpdates()
            } else {
                endRegularUpdates()
            }
        }
    }
    
    @Published
    private(set) var deviceInfo: DeviceInfo? {
        didSet {
            updateDeviceTimeIfNeeded()
            collectRecordedData()
        }
    }
    
    private var openRequests: [BluetoothRequest] = []
    
    private var runningRequest: BluetoothRequest?
    
    private var runningTransfer: TemperatureDataTransfer?
    
    // MARK: Regular updates
    
    func updateDeviceInfo() {
        guard case .configured = deviceState else {
            return
        }
        addRequest(.getInfo)
        
    }
    
    private var dataUpdateTimer: Timer?
    
    private func startRegularUpdates() {
        guard dataUpdateTimer == nil else {
            return
        }
        print("Starting updates")
        dataUpdateTimer = Timer.scheduledTimer(withTimeInterval: updateInterval, repeats: true) { [weak self] timer in
            guard let self = self else {
                timer.invalidate()
                return
            }
            self.updateDeviceInfo()
        }
        
        dataUpdateTimer?.fire()
    }

    private func endRegularUpdates() {
        guard let dataUpdateTimer else {
            return
        }
        dataUpdateTimer.invalidate()
        runningRequest = nil
        self.dataUpdateTimer = nil
        print("Ending updates")
    }
    
    // MARK: Requests
    
    private func performNextRequest() {
        guard runningRequest == nil else {
            return
        }
        guard !openRequests.isEmpty else {
            return
        }
        let next = openRequests.removeFirst()
        //print("Starting request \(next)")
        
        guard connection.send(next.serialized) else {
            print("Failed to start request \(next)")
            performNextRequest()
            return
        }
        runningRequest = next
    }
    
    func addRequest(_ request: BluetoothRequest) {
        // TODO: Check if request already exists
        openRequests.append(request)
        performNextRequest()
    }
    
    // MARK: Device time
    
    private func updateDeviceTimeIfNeeded() {
        guard let deviceInfo else {
            return
        }
        guard !deviceInfo.hasDeviceStartTimeSet || deviceInfo.clockOffset > minimumOffsetToUpdateDeviceClock else {
            return
        }

        guard !openRequests.contains(where: { if case .setDeviceStartTime = $0 { return true }; return false }) else {
            return
        }
        let time = deviceInfo.deviceStartTime.seconds
        addRequest(.setDeviceStartTime(deviceStartTimeSeconds: time))
        print("Setting device start time to \(time) s (\(Date().seconds) current)")
    }
    
    // MARK: Data transfer
    
    @discardableResult
    func collectRecordedData() -> Bool {
        guard runningTransfer == nil else {
            return false
        }
        guard !openRequests.contains(where: { if case .getRecordingData = $0 { return true }; return false }) else {
            return false
        }
        guard let info = deviceInfo else {
            return false
        }
        guard info.numberOfStoredMeasurements > 0 else {
            return false
        }
        
        let transfer = TemperatureDataTransfer(info: info)
        runningTransfer = transfer
        let next = transfer.nextRequest()
        addRequest(next)
        return true
    }
    
    private func didReceive(data: Data, offset: Int, count: Int) {
        guard let runningTransfer else {
            return // TODO: Start new transfer?
        }
        runningTransfer.add(data: data, offset: offset, count: count)
        continueTransfer()
    }
    
    private func continueTransfer() {
        guard let runningTransfer else {
            return // TODO: Start new transfer?
        }
        let next = runningTransfer.nextRequest()
        addRequest(next)
    }
    
    private func decode(info: Data) {
        guard let newInfo = try? DeviceInfo(info: info) else {
            return
        }
        self.deviceInfo = newInfo
        if let runningTransfer {
            runningTransfer.update(info: newInfo)
            let next = runningTransfer.nextRequest()
            addRequest(next)
        } else if newInfo.numberOfStoredMeasurements > 0 {
            collectRecordedData()
        }
    }
}

extension BluetoothClient: DeviceManagerDelegate {
    
    func deviceManager(didReceive data: Data) {
        defer {
            performNextRequest()
        }
        guard let runningRequest else {
            print("No request active, but \(data) received")
            return
        }
        self.runningRequest = nil
        
        guard data.count > 0 else {
            print("No response data for request \(runningRequest)")
            return
        }
        
        guard let type = BluetoothResponseType(rawValue: data[0]) else {
            print("Unknown response \(data[0]) for request \(runningRequest)")
            return
        }
        switch type {
        case .success:
            break
        case .responseInProgress:
            // Retry the request
            addRequest(runningRequest)
            return
        default:
            print("Unknown response \(data[0]) for request \(runningRequest)")
            // If clearing the recording buffer fails due to byte mismatch,
            // then requesting new info will resolve the mismatch, and the transfer will be resumed
            // If requesting bytes fails due to the response size,
            // then requesting new info will update the response size, and the transfer will be resumed
            addRequest(.getInfo)
            return
        
        }
        let payload = data.dropFirst()
        
        switch runningRequest {
        case .getInfo:
            decode(info: payload)
        case .getRecordingData(let offset, let count):
            didReceive(data: payload, offset: offset, count: count)
        case .clearRecordingBuffer:
            didClearDeviceStorage()

        case .setDeviceStartTime:
            print("Device time set")
            break
        }
    }

    private func didClearDeviceStorage() {
        guard let runningTransfer else {
            return
        }
        runningTransfer.completeTransfer()
        storage.add(runningTransfer.measurements)
        self.runningTransfer = nil

        updateDeviceTimeIfNeeded()
    }
    
    func deviceManager(didChangeState state: DeviceState) {
        DispatchQueue.main.async {
            self.deviceState = state
        }
    }
}

